/* ### CHECK-ME-BEFORE-RELEASE ### */
/*
 * Title:   xsound.cpp
 * Created: Spring 2000
 * Authors: Markus Noga
 *     $Id: xsound.cpp,v 1.3 2003/11/19 10:37:31 weyrich Exp $
 *
 * Copyright (c) 2000--2003
 * Gilbert Baumann <unk6@stud.uni-karlsruhe.de>
 * Markus Noga <markus@noga.de>
 * Tim Weyrich <weyrich@inf.ethz.ch>
 *
 * This file is part of the Pointshop3D system.
 * See http: *www.pointshop3d.com/ for more information.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307 USA
 *
 * Contact info@pointshop3d.com if any conditions of this
 * licensing are not clear to you.
 *
 * ---------------------------------------------------------------
 *
 * $Log: xsound.cpp,v $
 * Revision 1.3  2003/11/19 10:37:31  weyrich
 * added Pointshop3D headers
 *
 *
 */

#include <pisound.h>

// Disable Code On SGIs:
#ifndef _SGIAPI

#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/soundcard.h>
#include <errno.h>

#define DSP_PATH    "/dev/dsp"

//! Sound thread.
static void *xsound_thread(void *arg) {
  XSound *snd=(XSound*) arg;

  while(1) {
    snd->syncBuffer();
    snd->renderAndPlayBuffer();
  }

  return NULL;  // never reached.
}

//! Create an XSound object.
//
XSound::XSound(const char *dev,
               uint32 ra, uint32 bi, uint32 ch,
               uint32 bufsz, uint32 bufs)
  : IfSound(dev,ra,bi,ch,bufsz,bufs),
    pth(0)
{
  uint32 size,stereo,frag;
  uint32 i;
  audio_buf_info oinfo;

  // check superclass initialization validity
  //
  if(!isValid())
    return;
  valid=false;

  // default device name
  //
  if(!deviceName)
    deviceName=strdup(DSP_PATH);

  fd=open(deviceName, O_WRONLY | O_BINARY);
  if(fd<0) {
    genError("opening",errno);
    return;
  }

  // check for capabilities
  //
  if(ioctl(fd,SNDCTL_DSP_GETCAPS,&i)<0 || !(i&DSP_CAP_TRIGGER)) {
    genError("query trigger",errno);
    close(fd);
    return;
  }

  // set application profile (no matter if it fails).
  //
  i=APF_CPUINTENS;
  ioctl(fd,SNDCTL_DSP_PROFILE,&i);

  // set sample size
  //
  i=size= bits==8 ? AFMT_U8 : AFMT_S16_LE;
  if(ioctl(fd,SNDCTL_DSP_SETFMT,&i)<0 || i!=size) {
    genError("set format",errno);
    close(fd);
    return;
  }

  // set channels
  //
  i=stereo=channels>1 ? 1 : 0;
  if(ioctl(fd,SNDCTL_DSP_STEREO,&i)<0 || i!=stereo) {
    genError("set channels",errno);
    close(fd);
    return;
  }

  // set sample rate
  // (may change value, to be passed on later!)
  //
  if(ioctl(fd,SNDCTL_DSP_SPEED,&rate)<0) {
    genError("set rate",errno);
    close(fd);
    return;
  }

  // set & validate buffering
  //
  frag=0;
  bufSize*=getSampleSize();
  while(bufSize>1) {
    frag++;
    bufSize >>=1;
  }
  frag|=buffers<<16;

  if(ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &frag)<0) {
    genError("set buffering",errno);
    close(fd);
    return;
  }

  if(ioctl(fd, SNDCTL_DSP_GETOSPACE, &oinfo)<0) {
    genError("query buffering",errno);
    close(fd);
    return;
  }
  bufSize=oinfo.fragsize/getSampleSize();
  buffers=oinfo.fragstotal;

  bufdata=new uint8[getBufSizeBytes()];

  if(bufdata)
    valid  =true;
}

XSound::~XSound() {
  if(isValid()) {
    stop();
    delete bufdata;
    close(fd);
  }
}

void XSound::play() {
  // pre-fill output queue
  //
  if(isValid()) {
    setEnable(false);
    do
      renderAndPlayBuffer();
    while(getNonBlocking()>=getBufSizeBytes());
    setEnable(true);

    if(pithread_create(&pth, NULL, xsound_thread, this)) {
      // FIXME: error
    }
  }
}

void XSound::stop() {
  if(isValid() && pth) {
    if(pithread_cancel(pth)) {
      // FIXME: error
    }
    pth=0;
  }
}

void XSound::renderAndPlayBuffer() {
  renderBuffer(bufdata,getBufSize());
  write(fd,bufdata,getBufSizeBytes());
}

void XSound::syncBuffer() {
 fd_set writeset;

 FD_ZERO(&writeset);
 FD_SET(fd,&writeset);
 select(fd+1,NULL,&writeset,NULL,NULL);
}

void XSound::setEnable(bool enable) {
  uint32 mode= enable ? PCM_ENABLE_OUTPUT : 0;
  uint32    i=mode;
  if(ioctl(fd,SNDCTL_DSP_SETTRIGGER,&i)<0 || i!=mode) {
    // FIXME: error "cannot trigger mode"
  }
}

uint32 XSound::getNonBlocking() {
  audio_buf_info info;
  ioctl(fd, SNDCTL_DSP_GETOSPACE, &info);
  return info.bytes;
}

void XSound::reset() {
  ioctl(fd, SNDCTL_DSP_RESET);
}

void XSound::genError(const char *msg,int nr) {
  if(nr) {
    char  *err=strerror(nr);
    uint32 len=strlen(msg)+strlen(err)+10;

    errMsg=new char[len];
    if(errMsg)
      sprintf(errMsg,"sound: %s: %s",msg,err);
  } else
    IfSound::genError(msg);
}

#endif // ifndef _SGIAPI
