// ### CHECK-ME-BEFORE-RELEASE ###
//
// Title:   outstream.cpp
// Created: Thu Apr 18 14:31:26 2002
// Authors: Tim Weyrich <weyrich@inf.ethz.ch>
//
// Copyright (c) 2001--2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//
// ---------------------------------------------------------------
//
// $Log: outstream.cpp,v $
// Revision 1.9  2003/11/19 13:04:02  weyrich
// added Pointshop3D headers
//
// Revision 1.8  2003/11/17 15:05:52  weyrich
// new API allowing for more than 32 surfel properties
//
// Revision 1.7  2003/03/20 14:52:52  weyrich
// smaller fixes of implicit coordinate system transformations
//
// Revision 1.6  2002/12/18 17:05:45  weyrich
// added feature relations support (experimental)
//
// Revision 1.5  2002/11/18 13:09:12  weyrich
// added markers and tangential vectors
//
// Revision 1.4  2002/06/04 16:03:30  weyrich
// minor fixes
//
// Revision 1.3  2002/05/23 13:59:56  weyrich
// First official release.
//
// Revision 1.2  2002/05/17 13:58:16  weyrich
// PointShop3D integration nearly finished
//
// Revision 1.1.1.1  2002/05/08 08:58:47  weyrich
// surfel file format
//

#include <pisystem.h>  // for alloca
#include <pistring.h>

#include "api.h"
#include "internal/sflmat.h"

using namespace sfl;

/* KNOWN BUGS:
 *
 *  Dumping surfel-set-attributes during a resolution is not prevented.
 */

///////////////////////////////////////////////////////////////////////////////

OutStream::OutStream()
  : wroteHeader(false),
    wroteSurfelSetList(false),
    inHeader(false),
    inSurfelSetList(false),
    inSurfelSet(false),
    inResolution(false),
    numAnnouncedSurfelSets(0),
    gffDepth(0)
{
}

OutStream::~OutStream()
{
  if (inResolution)     endResolution();
  if (inSurfelSet)      endSurfelSet();
  if (inSurfelSetList)  endSurfelSetList();
}

///////////////////////////////////////////////////////////////////////////////

void  OutStream::assertHeader(void)
{
  if (!wroteHeader)
    {
      if (inHeader)
        fprintf(stderr, "%s:%d: WARNING: Forgot OutStream::endHeader()",
                __FILE__, __LINE__);
      else
        beginHeader();
      endHeader();
    }
}

void  OutStream::assertInSurfelSetList(void)
{
  assertHeader();
  if (!inSurfelSetList)
    {
      fprintf(stderr, "%s:%d: WARNING: Forgot OutStream::endSurfelSetList()",
              __FILE__, __LINE__);
      beginSurfelSetList(1);
    }
}

///////////////////////////////////////////////////////////////////////////////

OutStream  *OutStream::open(const char *fName, const char *mode)
{
  OutStream  *result = new OutStream();
  char  *m = (char *)alloca(strlen(mode)+2);

  strcpy(m, mode);
  strcat(m, "w");

  if ((result->gff = gff_open(fName, m, SFL_MAGIC, SFL_VERSION)))
    return result;

  delete result;
  return NULL;
}

int  OutStream::close(OutStream *out)
{
  delete out;
  return 0;
}

///////////////////////////////////////////////////////////////////////////////

int  OutStream::beginHeader(void)
{
  //fprintf(stderr, "[beginHeader]\n");
  if (!wroteHeader)
    {
      inHeader = true;
      return 0;
    }
  else
    {
      fprintf(stderr, "%s:%d: Already wrote header.\n", __FILE__, __LINE__);
      return 1;
    }
}

int  OutStream::endHeader(void)
{
  //fprintf(stderr, "[endHeader]\n");
  if (inHeader)
    {
      gff_write_field(gff, SFLTAG_SURFEL_SET_LIST_IDX, GFFTYPE_UINT32, 1, 0);

      gff_write_downlink(gff, 1);       // Down-link to surfel set list.
      wroteHeader = true;
      inHeader = false;

      return 0;
    }
  else
    {
      fprintf(stderr, "%s:%d:OutStream::endHeader(): WARNING: Not in header.\n",
              __FILE__, __LINE__);
      return 1;
    }
}

#define  ASSERT_HEADER_NOT_YET_WRITTEN(s)                       \
do {                                                            \
  if (wroteHeader)                                              \
    {                                                           \
      fprintf(stderr,                                           \
              "%s:%d: WARNING: Can not write field %s, "        \
              "since SFL header has already been dumped.\n",    \
              __FILE__, __LINE__, s);                           \
      return *this;                                             \
    }                                                           \
} while(0)

OutStream  &OutStream::setTitle(const char *title)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("TITLE");
  gff_write_field_cstring(gff, SFLTAG_TITLE, title);
  return *this;
}

OutStream  &OutStream::setAuthor(const char *author)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("AUTHOR");
  gff_write_field_cstring(gff, SFLTAG_AUTHOR, author);
  return *this;
}

OutStream  &OutStream::setApplication(const char *appl)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("APPLICATION");
  gff_write_field_cstring(gff, SFLTAG_APPLICATION, appl);
  return *this;
}

OutStream  &OutStream::setDescription(const char *descr)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("DESCRIPTION");
  gff_write_field_cstring(gff, SFLTAG_DESCRIPTION, descr);
  return *this;
}

OutStream  &OutStream::setVersion(uint32 version)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("VERSION");
  gff_write_field(gff, SFLTAG_VERSION, GFFTYPE_UINT32, 1, version);
  return *this;
}

OutStream  &OutStream::setDate(const char *date)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("DATE");
  gff_write_field_cstring(gff, SFLTAG_DATE, date);
  return *this;
}

OutStream  &OutStream::setRightVec3(double x, double y, double z)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("RIGHT_VEC");
  gff_write_field(gff, SFLTAG_RIGHT_VEC, GFFTYPE_FLOAT, 3, x, y, z);
  return *this;
}

OutStream  &OutStream::setRightVec3(const float *vec)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("RIGHT_VEC");
  gff_write_field(gff, SFLTAG_RIGHT_VEC, GFFTYPE_FLOAT, 3, vec[0], vec[1], vec[2]);
  return *this;
}

OutStream  &OutStream::setRightVec3(const double *vec)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("RIGHT_VEC");
  gff_write_field(gff, SFLTAG_RIGHT_VEC, GFFTYPE_FLOAT, 3, vec[0], vec[1], vec[2]);
  return *this;
}

OutStream  &OutStream::setUpVec3(double x, double y, double z)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("UP_VEC");
  gff_write_field(gff, SFLTAG_UP_VEC, GFFTYPE_FLOAT, 3, x, y, z);
  return *this;
}

OutStream  &OutStream::setUpVec3(const float *vec)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("UP_VEC");
  gff_write_field(gff, SFLTAG_UP_VEC, GFFTYPE_FLOAT, 3, vec[0], vec[1], vec[2]);
  return *this;
}

OutStream  &OutStream::setUpVec3(const double *vec)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("UP_VEC");
  gff_write_field(gff, SFLTAG_UP_VEC, GFFTYPE_FLOAT, 3, vec[0], vec[1], vec[2]);
  return *this;
}

OutStream  &OutStream::setBackVec3(double x, double y, double z)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("BACK_VEC");
  gff_write_field(gff, SFLTAG_BACK_VEC, GFFTYPE_FLOAT, 3, x, y, z);
  return *this;
}

OutStream  &OutStream::setBackVec3(const float *vec)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("BACK_VEC");
  gff_write_field(gff, SFLTAG_BACK_VEC, GFFTYPE_FLOAT, 3, vec[0], vec[1], vec[2]);
  return *this;
}

OutStream  &OutStream::setBackVec3(const double *vec)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("BACK_VEC");
  gff_write_field(gff, SFLTAG_BACK_VEC, GFFTYPE_FLOAT, 3, vec[0], vec[1], vec[2]);
  return *this;
}

OutStream  &OutStream::setDefaultUnits(double units)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("DEFAULT_UNITS");
  gff_write_field(gff, SFLTAG_DEFAULT_UNITS, GFFTYPE_FLOAT, 1, units);
  return *this;
}

OutStream  &OutStream::setOverAllBoundingBox(double x1, double y1, double z1,
                                             double x2, double y2, double z2)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("BOUNDING_BOX");
  gff_write_field(gff, SFLTAG_BOUNDING_BOX, GFFTYPE_FLOAT, 6, x1, y1, z1, x2, y2, z2);
  return *this;
}

OutStream  &OutStream::setSceneCameraTransform4x4(const float *mat)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("SCENE_CAMERA_TRANSFORM");
  float  temp[16];
  apiSwapMat4x4ToTemp4x4(temp, mat);
  gff_write_field_arr(gff, SFLTAG_SCENE_CAMERA_TRANSFORM, GFFTYPE_FLOAT, 16, temp);
  return *this;
}

OutStream  &OutStream::setSceneCameraTransform4x4(const double *mat)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("SCENE_CAMERA_TRANSFORM");
  double  temp[16];
  apiSwapMat4x4ToTemp4x4(temp, mat);
  gff_write_field(gff, SFLTAG_SCENE_CAMERA_TRANSFORM, GFFTYPE_FLOAT, 16,
                  (float)temp[ 0], (float)temp[ 1], (float)temp[ 2], (float)temp[ 3],
                  (float)temp[ 4], (float)temp[ 5], (float)temp[ 6], (float)temp[ 7],
                  (float)temp[ 8], (float)temp[ 9], (float)temp[10], (float)temp[11],
                  (float)temp[12], (float)temp[13], (float)temp[14], (float)temp[15]);
  return *this;
}

OutStream  &OutStream::setSceneCameraTransformTwice4x4(const float *matSecond, const float *matFirst)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("SCENE_CAMERA_TRANSFORM");
  float  tempSec[16];
  float  tempFst[16];
  apiSwapMat4x4ToTemp4x4(tempSec, matSecond);
  apiSwapMat4x4ToTemp4x4(tempFst, matFirst);
  sflMat::mulMat16Mat16(tempSec, tempSec, tempFst);
  gff_write_field_arr(gff, SFLTAG_SCENE_CAMERA_TRANSFORM, GFFTYPE_FLOAT, 16, tempSec);
  return *this;
}

OutStream  &OutStream::setSceneCameraTransformTwice4x4(const double *matSecond, const double *matFirst)
{
  ASSERT_HEADER_NOT_YET_WRITTEN("SCENE_CAMERA_TRANSFORM");
  double  tempSec[16];
  double  tempFst[16];
  apiSwapMat4x4ToTemp4x4(tempSec, matSecond);
  apiSwapMat4x4ToTemp4x4(tempFst, matFirst);
  sflMat::mulMat16Mat16(tempSec, tempSec, tempFst);
  gff_write_field(gff, SFLTAG_SCENE_CAMERA_TRANSFORM, GFFTYPE_FLOAT, 16,
                  (float)tempSec[ 0], (float)tempSec[ 1], (float)tempSec[ 2], (float)tempSec[ 3],
                  (float)tempSec[ 4], (float)tempSec[ 5], (float)tempSec[ 6], (float)tempSec[ 7],
                  (float)tempSec[ 8], (float)tempSec[ 9], (float)tempSec[10], (float)tempSec[11],
                  (float)tempSec[12], (float)tempSec[13], (float)tempSec[14], (float)tempSec[15]);
  return *this;
}

///////////////////////////////////////////////////////////////////////////////


int  OutStream::beginSurfelSetList(uint32 numSurfelSets)
{
  //fprintf(stderr, "[beginSurfelSetList]\n");
  if (!inSurfelSetList)
    {
      if (!wroteSurfelSetList)
        {
          assertHeader();

          if (!gff_write_field(gff, SFLTAG_NUM_SURFEL_SETS, GFFTYPE_UINT32, 1, numSurfelSets))
            {
              numAnnouncedSurfelSets = numSurfelSets;
              numWrittenSurfelSets = 0;
              inSurfelSetList = true;

              return 0;
            }
        }
      else
        fprintf(stderr, "%s:%d: Already wrote surfel set list\n",
                __FILE__, __LINE__);
    }
  else
    fprintf(stderr, "%s:%d: Already in SurfelSetList\n",
            __FILE__, __LINE__);
  return 1;
}

int  OutStream::endSurfelSetList(void)
{
  //fprintf(stderr, "[endSurfelSetList]\n");
  if (inSurfelSetList)
    {
      if (numWrittenSurfelSets > 0)
        gff_write_uplink(gff);  // up-link from surfel sets

      gff_write_uplink(gff);    // up-link from surfel set list

      wroteSurfelSetList = true;
      inSurfelSetList = false;

      return 0;
    }
  else
    {
      fprintf(stderr, "%s:%d:OutStream::endSurfelSetList(): WARNING: Not in surfel set list.\n",
              __FILE__, __LINE__);
      return 1;
    }
}

///////////////////////////////////////////////////////////////////////////////

int  OutStream::beginSurfelSet(unsigned propertyMask, ssize_t numResolutions)
{
  return beginSurfelSet(PropMaskType(propertyMask), numResolutions);
}

int  OutStream::beginSurfelSet(const PropMaskType &propertyMask, ssize_t numResolutions)
{
  //fprintf(stderr, "[beginSurfelSet 0x%08x, %d]\n", propertyMask, (int)numResolutions);
  assertInSurfelSetList();
  if (!inSurfelSet)
    {
      if (numWrittenSurfelSets < numAnnouncedSurfelSets)
        {
          int  err=0;

          surfelBuffer.setPropertyMask(propertyMask);

          if (numWrittenSurfelSets == 0)
            err = gff_write_downlink(gff, numAnnouncedSurfelSets); // down-link to surfel sets

          if (!err &&
              !(err = gff_write_field(gff, SFLTAG_SURFEL_SET_NUM_RESOLUTIONS, GFFTYPE_UINT32, 1,
                                      (uint32)numResolutions)) &&
              !(err = gff_write_field(gff, SFLTAG_SURFEL_SET_FORMAT_VERSION, GFFTYPE_UINT32, 1,
                                      (uint32)SFL_SURFEL_FORMAT_VERSION)) &&
              !(err = gff_write_field(gff, SFLTAG_SURFEL_SET_SURFEL_SIZE, GFFTYPE_UINT32, 1,
                                      (uint32)surfelBuffer.getDiskSize())) &&
              !(err = gff_write_field_arr(gff, SFLTAG_SURFEL_SET_PROPERTIES, GFFTYPE_UINT32,
                                          ((propertyMask.getNumBits()+31)/32),
                                          propertyMask.getData())))
            {
              numAnnouncedResolutions = numResolutions;
              numWrittenResolutions = 0;

              {
                int  i;
                int  n=0;  // number of properties per surfel
                for(i=0; i<(int)MAX_PROPERTY_BITS; i++)
                  if (propertyMask.isSet(i) && surfelBuffer.getPropertySize((PropertyBitEnum)i))
                    n++;

                uint16  *indices = (uint16 *)alloca(n * sizeof(uint16));
                uint32  *offsets = (uint32 *)alloca(n * sizeof(uint32));

                for (i=0, n=0; i<(int)MAX_PROPERTY_BITS; i++)
                  if (propertyMask.isSet(i) && surfelBuffer.getPropertySize((PropertyBitEnum)i))
                    {
                      indices[n] = i;
                      offsets[n] = surfelBuffer.getPropertyOffset((PropertyBitEnum)i);
                      n++;
                    }

                err = gff_write_field_arr(gff, SFLTAG_SURFEL_SET_PROP_INDICES, GFFTYPE_UINT16, n, indices);
                if (!err)
                  err = gff_write_field_arr(gff, SFLTAG_SURFEL_SET_PROP_OFFSETS, GFFTYPE_UINT32, n, offsets);
              }
            }

          if (!err)
            {
              inSurfelSet = true;
              return 0;
            }
          else
            fprintf(stderr, "%s:%d:OutStream::beginSurfelSet(): Write error.\n",
                    __FILE__, __LINE__);
        }
      else
        fprintf(stderr, "%s:%d:OutStream::beginSurfelSet(): Can not write any more surfel sets, since only %d were announced in the surfel set list.\n",
                __FILE__, __LINE__, numAnnouncedSurfelSets);
    }
  else
    fprintf(stderr, "%s:%d: Already in SurfelSetList\n",
            __FILE__, __LINE__);
  return 1;
}

#define  ASSERT_IN_SURFEL_SET(s)                                                \
do {                                                                            \
  if (!inSurfelSet)                                                             \
    {                                                                           \
      fprintf(stderr,                                                           \
              "%s:%d:OutStream::%s: Can not be called outside surfel set.\n",   \
              __FILE__, __LINE__, s);                                           \
      return 1;                                                                 \
    }                                                                           \
} while(0)

int  OutStream::setSurfelSetRes0Index(int idx)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetRes0Index");
  return gff_write_field(gff, SFLTAG_SURFEL_SET_RES0_IDX, GFFTYPE_UINT32, 1, idx);
}

int  OutStream::setSurfelSetUnits(double units)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetUnits");
  return gff_write_field(gff, SFLTAG_SURFEL_SET_UNITS, GFFTYPE_FLOAT, 1, (float)units);
}

int  OutStream::setSurfelSetTransform3x3(const float *mat)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetTransform3x3");
  float  temp[16];
  apiSwapMat4x4ToTemp4x4(temp, mat);
  return gff_write_field(gff, SFLTAG_SURFEL_SET_TRANSFORM, GFFTYPE_FLOAT, 16,
                         temp[0], temp[1], temp[2], 0.0f,
                         temp[3], temp[4], temp[5], 0.0f,
                         temp[6], temp[7], temp[8], 0.0f,
                         0.0f, 0.0f, 0.0f, 1.0f);
}

int  OutStream::setSurfelSetTransform4x4(const float *mat)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetTransform4x4");
  float  temp[16];
  apiSwapMat4x4ToTemp4x4(temp, mat);
  return gff_write_field_arr(gff, SFLTAG_SURFEL_SET_TRANSFORM, GFFTYPE_FLOAT, 16, temp);
}

int  OutStream::setSurfelSetTransform3x3(const double *mat)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetTransform3x3");
  double  temp[16];
  apiSwapMat4x4ToTemp4x4(temp, mat);
  return gff_write_field(gff, SFLTAG_SURFEL_SET_TRANSFORM, GFFTYPE_FLOAT, 16,
                         (float)temp[0], (float)temp[1], (float)temp[2], 0.0f,
                         (float)temp[3], (float)temp[4], (float)temp[5], 0.0f,
                         (float)temp[6], (float)temp[7], (float)temp[8], 0.0f,
                         0.0f, 0.0f, 0.0f, 1.0f);
}

int  OutStream::setSurfelSetTransform4x4(const double *mat)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetTransform4x4");
  double  temp[16];
  apiSwapMat4x4ToTemp4x4(temp, mat);
  return gff_write_field(gff, SFLTAG_SURFEL_SET_TRANSFORM, GFFTYPE_FLOAT, 16,
                         (float)temp[ 0], (float)temp[ 1], (float)temp[ 2], (float)temp[ 3],
                         (float)temp[ 4], (float)temp[ 5], (float)temp[ 6], (float)temp[ 7],
                         (float)temp[ 8], (float)temp[ 9], (float)temp[10], (float)temp[11],
                         (float)temp[12], (float)temp[13], (float)temp[14], (float)temp[15]);
}

int  OutStream::setSurfelSetTransformTwice4x4(const float *matSecond, const float *matFirst)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetTransformTwice4x4");
  float  tempSec[16];
  float  tempFst[16];
  apiSwapMat4x4ToTemp4x4(tempSec, matSecond);
  apiSwapMat4x4ToTemp4x4(tempFst, matFirst);
  sflMat::mulMat16Mat16(tempSec, tempSec, tempFst);
  return gff_write_field_arr(gff, SFLTAG_SURFEL_SET_TRANSFORM, GFFTYPE_FLOAT, 16, tempSec);
}

int  OutStream::setSurfelSetTransformTwice4x4(const double *matSecond, const double *matFirst)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetTransformTwice4x4");
  double  tempSec[16];
  double  tempFst[16];
  apiSwapMat4x4ToTemp4x4(tempSec, matSecond);
  apiSwapMat4x4ToTemp4x4(tempFst, matFirst);
  sflMat::mulMat16Mat16(tempSec, tempSec, tempFst);
  return gff_write_field(gff, SFLTAG_SURFEL_SET_TRANSFORM, GFFTYPE_FLOAT, 16,
                         (float)tempSec[ 0], (float)tempSec[ 1], (float)tempSec[ 2], (float)tempSec[ 3],
                         (float)tempSec[ 4], (float)tempSec[ 5], (float)tempSec[ 6], (float)tempSec[ 7],
                         (float)tempSec[ 8], (float)tempSec[ 9], (float)tempSec[10], (float)tempSec[11],
                         (float)tempSec[12], (float)tempSec[13], (float)tempSec[14], (float)tempSec[15]);
}

int  OutStream::setSurfelSetBoundingBox(double x1, double y1, double z1, double x2, double y2, double z2)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetBoundingBox");
  return gff_write_field(gff, SFLTAG_BOUNDING_BOX, GFFTYPE_FLOAT, 6,
                         (float)x1, (float)y1, (float)z1, (float)x2, (float)y2, (float)z2);
}

int  OutStream::setSurfelSetIdentifier(const char *ident)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetIdentifier");
  return gff_write_field_cstring(gff, SFLTAG_SURFEL_SET_IDENTIFIER, ident);
}

int  OutStream::setSurfelSetMarkerSurfelIndices(int n, const uint32 *idxvec)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetMarkerSurfelIndices");
  return gff_write_field_arr(gff, SFLTAG_SURFEL_SET_MARKER_IDXVEC, GFFTYPE_UINT32, n, idxvec);
}

int  OutStream::setSurfelSetMarkerSurfelUs(int n, const float *uvec)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetMarkerSurfelUs");
  return gff_write_field_arr(gff, SFLTAG_SURFEL_SET_MARKER_UVEC, GFFTYPE_FLOAT, n, uvec);
}

int  OutStream::setSurfelSetMarkerSurfelVs(int n, const float *vvec)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetMarkerSurfelVs");
  return gff_write_field_arr(gff, SFLTAG_SURFEL_SET_MARKER_VVEC, GFFTYPE_FLOAT, n, vvec);
}

int  OutStream::setSurfelSetExpFeatureRels(int nRels, const uint32 *fromToIdxVec)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetExpFeatureRels");
  return gff_write_field_arr(gff, SFLTAG_SURFEL_SET_EXP_FEATURE_RELS, GFFTYPE_UINT32, 2*nRels, fromToIdxVec);
}

int  OutStream::setSurfelSetSRFCalibrations(int nCalibs, const float *cops, const float *pinvs)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetSRFCalibrations");
  float  *userFloatVec = (float *)malloc(nCalibs * 100*12 * sizeof(float));
  int     i;
  for (i=0; i<3*nCalibs; i++)
    userFloatVec[i] = cops[i];
  for (i=0; i<9*nCalibs; i++)
    userFloatVec[3*nCalibs + i] = pinvs[i];
  int res = gff_write_field_arr(gff, SFLTAG_SURFEL_SET_SRF_CALIBFLOATS, GFFTYPE_FLOAT, 12*nCalibs, userFloatVec);
  free(userFloatVec);
  return res;
}

int  OutStream::setSurfelSetSRFCalibrations(int nCalibs, const double *cops, const double *pinvs)
{
  ASSERT_IN_SURFEL_SET("setSurfelSetSRFCalibrations");
  float  *userFloatVec = (float *)malloc(nCalibs * 100*12 * sizeof(float));
  int     i;
  for (i=0; i<3*nCalibs; i++)
    userFloatVec[i] = (float)cops[i];
  for (i=0; i<9*nCalibs; i++)
    userFloatVec[3*nCalibs + i] = (float)pinvs[i];
  int res = gff_write_field_arr(gff, SFLTAG_SURFEL_SET_SRF_CALIBFLOATS, GFFTYPE_FLOAT, 12*nCalibs, userFloatVec);
  free(userFloatVec);
  return res;
}

int  OutStream::endSurfelSet(void)
{
  //fprintf(stderr, "[endSurfelSet]\n");
  if (inSurfelSet)
    {
      if (numWrittenResolutions > 0)
        gff_write_uplink(gff);  // up-link from resolutions

      inSurfelSet = false;
      numWrittenSurfelSets++;

      return 0;
    }
  else
    {
      fprintf(stderr, "%s:%d:OutStream::endSurfelSet(): WARNING: Not in surfel set.\n",
              __FILE__, __LINE__);
      return 1;
    }
}

///////////////////////////////////////////////////////////////////////////////

int  OutStream::beginResolution(ssize_t numSurfels)
{
  //fprintf(stderr, "[beginResolution]\n");
  if (inSurfelSet)
    {
      if (numWrittenResolutions < numAnnouncedResolutions)
        {
          int  err=0;

          if (numWrittenResolutions == 0)
            err = gff_write_downlink(gff, numAnnouncedResolutions); // down-link to resolutions

          if (!err &&
              !gff_write_field(gff, SFLTAG_SURFEL_RSL_NUM_SURFELS, GFFTYPE_UINT32, 1, numSurfels) &&
              !gff_write_data_field(gff, SFLTAG_SURFEL_RSL_DATA, numSurfels * surfelBuffer.getDiskSize()))
            {
              numAnnouncedSurfels = numSurfels;
              numWrittenSurfels = 0;

              inResolution = true;

              return 0;
            }
          else
            fprintf(stderr, "%s%d:OutStream::beginSurfelSet(): Write error\n",
                    __FILE__, __LINE__);
        }
      else
        fprintf(stderr, "%s:%d:OutStream::beginSurfelSet(): Can not write any more resolutions, since only %d were announced for this surfel set.\n",
                __FILE__, __LINE__, numAnnouncedSurfelSets);
    }
  else
    fprintf(stderr, "%s:%d:OutStream::beginResolution(): Must not be called outside a surfel set.\n",
            __FILE__, __LINE__);
  return 1;
}

int  OutStream::endResolution(void)
{
  //fprintf(stderr, "[endResolution]\n");
  if (inResolution)
    {
      int  res = 0;
      if (numWrittenSurfels < numAnnouncedSurfels)
        {
          fprintf(stderr, "%s:%d:OutStream::endResolution(): Too few surfels were written. Filling up by repeating the last surfel...\n",
                  __FILE__, __LINE__);
          while (numWrittenSurfels < numAnnouncedSurfels)
            if (endSurfel())
              break;
          res = 1;
        }
      flushRwBuffer();

      inResolution = false;
      numWrittenResolutions++;

      return res;
    }
  else
    {
      fprintf(stderr, "%s:%d:OutStream::endResolution(): WARNING: Not in a resolution.\n",
              __FILE__, __LINE__);
      return 1;
    }
}

///////////////////////////////////////////////////////////////////////////////

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
